#include <cstring>
#include <iostream>
#include <stack>
#include <stdexcept>
#include <string>
#include <vector>

#include "ExprEval.h"

using namespace std;


namespace ExprEval
{

///////////////////////////////////////////////////////////////////////////////
//
// There are 16 precedence categories, some of which contain only one operator.
// Operators in the same category have equal precedence with each other.
//
// Where duplicates of operators appear in the table, the first occurrence is
// unary, the second binary. Each category has an associativity rule: left to
// right, or right to left. In the absence of parentheses, these rules resolve
// the grouping of expressions with operators of equal precedence.

// The precedence of each operator in the following table is indicated by its
// order in the table. The first category (on the first line) has the highest
// precedence. Operators on the same line have equal precedence.

// Operators                              Associativity
//
// () [] -> :: .                          left to right
// ! ~ + - ++ -- & * sizeof new delete    right to left
// .* ->*                                 left to right
// * / %                                  left to right
// + -                                    left to right
// << >>                                  left to right
// < <= > >=                              left to right
// == !=                                  left to right
// &                                      left to right
// ^                                      left to right
// |                                      left to right
// &&                                     left to right
// ||                                     right to left
// ?:                                     left to right
// = *= /= %= += -= &= ^= |= <<= >>=      right to left
// ,                                      left to right
//
///////////////////////////////////////////////////////////////////////////////

enum { SHL=1, SHR=2, LE=3, GE=4, EQ=5, NE=6, AND=7, OR=8, UPLUS=9, UMIN=10 };
static char priority[128] =
{
	0,11,11,10,10, 9, 9, 5,  4,15,15, 0, 0, 0, 0, 0, //  0:  << >> <= >= == != && || U+ U-
	0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0, // 16:
	0,15, 0, 0, 0,13, 8, 0,  0, 0,13,12, 1,12, 0,13, // 32:  !"#$%&' ()*+,-./
	0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 3, 0,10, 0,10, 3, // 48: 01234567 89:;<=>?
	0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0, // 64: @ABCDEFG HIJKLMNO
	0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0,16, 0,16, 7, 0, // 80: PQRSTUVW XYZ[\]^_
	0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0, // 96: `abcdefg hijklmno
	0, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 6, 0,15, 0  //112: pqrstuvw xyz{|}~
};


inline void skipws(char *&src)
{
	while (*src == ' ' || *src == '\t')
		++src;
}

inline bool isinteger(char *&src, int &integer)
{
	if (!isdigit(*src))
		return false;
	integer = 0;
	while (isdigit(*src))
		integer = *src++ - '0' + 10 * integer;
	return true;
}

inline bool isword(char *&src, string &word)
{
	if (!isalpha(*src) && *src != '_')
		return false;
	char *p = src + 1;
	while (isalnum(*p) || *p == '_')
		++p;
	word = string();
	word . append(src, p - src);
	src = p;
	return true;
}

inline bool isoperator(char *&src, int &op, bool bStartExpr)
{
	if (!strchr("[]!~*/%+-<=>&^|?:,", *src))
		return false;
	// What if operator not recognized ...
	op = 0;
	switch (*src)
	{
		case '|': if (src[1] == '|') op = OR; break;
		case '&': if (src[1] == '&') op = AND; break;
		case '=': if (src[1] == '=') op = EQ; break;
		case '!': if (src[1] == '=') op = NE; break;
		case '<': if (src[1] == '<') op = SHL; else if (src[1] == '=') op = LE; break;
		case '>': if (src[1] == '>') op = SHR; else if (src[1] == '=') op = GE; break;
	}
	if (op != 0)
	{
		src += 2;
		return true;
	}
	if (priority[*src] > 0)
	{
		if (bStartExpr && *src == '+')
			op = UPLUS;
		else if (bStartExpr && *src == '-')
			op = UMIN;
		else
			op = *src;
		++src;
		return true;
	}
	throw expr_exception("Bad operator");
}

inline bool isrightgrouping(int op)
{
   return (op == '!' || op == '~' || op == UPLUS || op == UMIN || op == OR);
}


static void evaluate_operator(stack<t_value,vector<t_value> > &value_stack, int op)
{
	switch (op)
	{
		case '?': return;
      case '[': return;
	}
	// What if value_stack empty ...
	if (value_stack . empty())
		throw expr_exception("Unbalanced expression");
	t_value arg1 = value_stack . top();
	value_stack . pop();
	switch (op)
	{
		case '!': value_stack . push(!arg1); return;
		case '~': value_stack . push(~(int)arg1); return;
		case UPLUS: value_stack . push(arg1); return;
		case UMIN: value_stack . push(-arg1); return;
	}
	// What if value_stack empty ...
	if (value_stack . empty())
		throw expr_exception("Unbalanced expression");
	t_value arg2 = value_stack . top();
	value_stack . pop();
	switch (op)
	{
      case ']': value_stack . push(0); return;
		case '*': value_stack . push(arg2 * arg1); return;
		case '/':
			// What if division by zero ...
			if (arg1 == 0)
				throw expr_exception("Division by zero");
			value_stack . push(arg2 / arg1);
			return;
		case '%':
			// What if mod by zero ...
			if ((int)arg1 == 0)
				throw expr_exception("Modulo by zero");
			value_stack . push((int)arg2 % (int)arg1);
			return;
		case '+': value_stack . push(arg2 + arg1); return;
		case '-': value_stack . push(arg2 - arg1); return;
		case SHL: value_stack . push(arg2 << arg1); return;
		case SHR: value_stack . push(arg2 >> arg1); return;
		case '<': value_stack . push(arg2 < arg1); return;
		case LE:  value_stack . push(arg2 <= arg1); return;
		case '>': value_stack . push(arg2 > arg1); return;
		case GE:  value_stack . push(arg2 >= arg1); return;
		case EQ:  value_stack . push(arg2 == arg1); return;
		case NE:  value_stack . push(arg2 != arg1); return;
		case '&': value_stack . push((int)arg2 & (int)arg1); return;
		case '^': value_stack . push((int)arg2 ^ (int)arg1); return;
		case '|': value_stack . push((int)arg2 | (int)arg1); return;
		case AND: value_stack . push(arg2 && arg1); return;
		case OR:  value_stack . push(arg2 || arg1); return;
		case ',': value_stack . push(arg1); return;
	}
	if (value_stack . empty())
		throw expr_exception("Unbalanced expression");
	t_value arg3 = value_stack . top();
	value_stack . pop();
	switch (op)
	{
		case ':': value_stack . push(arg3 ? arg2 : arg1); return;
	}
	// What if operator not recognized ...
	throw expr_exception("Unrecognized operator");
}

} // namespace ExprEval

void eval_expression(const string &expr, t_value &result, t_value (*get_variable)(const char *name))
{
	using namespace ExprEval;

	stack<t_value,vector<t_value> > value_stack;
	stack<int,vector<int> > op_stack;
	char *src = const_cast<char *>(expr . c_str());
	int integer;
	string word;
	int op;
	bool bStartExpr = true; // For handling unary + and -

//value_stack . push(0); value_stack . pop();

	skipws(src);
	while (*src)
	{
		if (isinteger(src, integer))
		{
			value_stack . push(integer);
			bStartExpr = false;
		}
		else if (get_variable && isword(src, word))
		{
			value_stack . push((*get_variable)(word . data()));
			bStartExpr = false;
		}
		else if (isoperator(src, op, bStartExpr))
		{
			while (!op_stack . empty()
            && (priority[op] < priority[op_stack . top()]
                || !isrightgrouping(op)
                   && priority[op] == priority[op_stack . top()]))
			{
				evaluate_operator(value_stack, op_stack . top());
				op_stack . pop();
			}
			op_stack . push(op);
			bStartExpr = true;
		}
		else if (*src == '(')
		{
			op_stack . push('(');
			++src;
		}
		else if (*src == ')')
		{
			// What if op_stack is or becomes empty ...
			while (!op_stack . empty() && op_stack . top() != '(')
			{
				evaluate_operator(value_stack, op_stack . top());
				op_stack . pop();
			}
			if (op_stack . empty() || op_stack . top() != '(')
				throw expr_exception("Right parenthesis without matching left parenthesis");
			op_stack . pop();
			++src;
		}
		else
		{
			throw expr_exception("Unrecognized token");
		}
		skipws(src);
	}
	while (!op_stack . empty())
	{
		evaluate_operator(value_stack, op_stack . top());
		op_stack . pop();
	}
	// What if value_stack does not contain exactly one value ...
	if (value_stack . size() != 1)
		throw expr_exception("Unbalanced expression");
	result = value_stack . top();
}


